/*
	SDL - Simple DirectMedia Layer
	Copyright (C) 1997, 1998, 1999, 2000  Sam Lantinga

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Library General Public
	License as published by the Free Software Foundation; either
	version 2 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Library General Public License for more details.

	You should have received a copy of the GNU Library General Public
	License along with this library; if not, write to the Free
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Sam Lantinga
	slouken@devolution.com
*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_fbevents.c,v 1.5.2.5 2000/05/02 23:09:42 hercules Exp $";
#endif

/*
 * Microwindows based SDL event driver implementation.
 *
 * Copyright (c) 2001 by Greg Haerr <greg@centurysoftware.com>
 *
 * Handle the event stream, converting Microwindows events into SDL events 
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "SDL.h"
#include "SDL_mutex.h"
#include "SDL_sysevents.h"
#include "SDL_sysvideo.h"
#include "SDL_events_c.h"
#include "SDL_fbvideo.h"
#include "SDL_fbevents_c.h"

/* The translation table from a Microwindows keysym to an SDL keysym */
static SDLKey MISC_keymap[256];
static SDL_keysym *FB_TranslateKey(int ch, int scancode, SDL_keysym *keysym);

#if 0
void FB_vgamousecallback(int button, int dx, int dy)
{
	int button_1, button_3;
	int button_state;
	int state_changed;
	int i;
	Uint8 state;

	if ( dx || dy ) {
		posted += SDL_PrivateMouseMotion(0, 1, dx, dy);
	}

	/* Swap button 1 and 3 */
	button_1 = (button & 0x04) >> 2;
	button_3 = (button & 0x01) << 2;
	button &= ~0x05;
	button |= (button_1|button_3);

	/* See what changed */
	button_state = SDL_GetMouseState(NULL, NULL);
	state_changed = button_state ^ button;
	for ( i=0; i<8; ++i ) {
		if ( state_changed & (1<<i) ) {
			if ( button & (1<<i) ) {
				state = SDL_PRESSED;
			} else {
				state = SDL_RELEASED;
			}
			posted += SDL_PrivateMouseButton(state, i+1, 0, 0);
		}
	}
}

/* Handle switching to another VC, returns when our VC is back */
static void switch_vt(_THIS, unsigned short which)
{
	struct vt_stat vtstate;
	unsigned short current;
	SDL_Surface *screen;
	__u16 saved_pal[3*256];
	Uint32 screen_arealen;
	Uint8 *screen_contents;

	/* Figure out whether or not we're switching to a new console */
	if ( (ioctl(keyboard_fd, VT_GETSTATE, &vtstate) < 0) ||
	     (which == vtstate.v_active) ) {
		return;
	}
	current = vtstate.v_active;

	/* Save the contents of the screen, and go to text mode */
	SDL_mutexP(hw_lock);
	screen = SDL_VideoSurface;
	screen_arealen = (screen->h*screen->pitch);
	screen_contents = (Uint8 *)malloc(screen_arealen);
	if ( screen_contents ) {
		memcpy(screen_contents, screen->pixels, screen_arealen);
	}
	FB_SavePaletteTo(this, 256, saved_pal);
	ioctl(keyboard_fd, KDSETMODE, KD_TEXT);

	/* New console, switch to it */
	if ( ioctl(keyboard_fd, VT_ACTIVATE, which) == 0 ) {
		/* Wait for our console to be activated again */
		ioctl(keyboard_fd, VT_WAITACTIVE, which);
		while ( ioctl(keyboard_fd, VT_WAITACTIVE, current) < 0 ) {
			if ( (errno != EINTR) && (errno != EAGAIN) ) {
				/* Unknown VT error - cancel this */
				break;
			}
			SDL_Delay(500);
		}
	}

	/* Restore graphics mode and the contents of the screen */
	ioctl(keyboard_fd, KDSETMODE, KD_GRAPHICS);
	FB_RestorePaletteFrom(this, 256, saved_pal);
	if ( screen_contents ) {
		memcpy(screen->pixels, screen_contents, screen_arealen);
		free(screen_contents);
	}
	SDL_mutexV(hw_lock);
}
#endif

void FB_InitOSKeymap(_THIS)
{
	int i;

	/* Map the miscellaneous keys */
	for ( i=0; i<SDL_TABLESIZE(MISC_keymap); ++i )
		MISC_keymap[i] = SDLK_UNKNOWN;

	/* These keysyms have 0xF8 as the high byte */
	MISC_keymap[MWKEY_KP0&0xFF] = SDLK_KP0;
	MISC_keymap[MWKEY_KP1&0xFF] = SDLK_KP1;
	MISC_keymap[MWKEY_KP2&0xFF] = SDLK_KP2;
	MISC_keymap[MWKEY_KP3&0xFF] = SDLK_KP3;
	MISC_keymap[MWKEY_KP4&0xFF] = SDLK_KP4;
	MISC_keymap[MWKEY_KP5&0xFF] = SDLK_KP5;
	MISC_keymap[MWKEY_KP6&0xFF] = SDLK_KP6;
	MISC_keymap[MWKEY_KP7&0xFF] = SDLK_KP7;
	MISC_keymap[MWKEY_KP8&0xFF] = SDLK_KP8;
	MISC_keymap[MWKEY_KP9&0xFF] = SDLK_KP9;
	MISC_keymap[MWKEY_KP_PERIOD&0xFF] = SDLK_KP_PERIOD;
	MISC_keymap[MWKEY_KP_DIVIDE&0xFF] = SDLK_KP_DIVIDE;
	MISC_keymap[MWKEY_KP_MULTIPLY&0xFF] = SDLK_KP_MULTIPLY;
	MISC_keymap[MWKEY_KP_MINUS&0xFF] = SDLK_KP_MINUS;
	MISC_keymap[MWKEY_KP_PLUS&0xFF] = SDLK_KP_PLUS;
	MISC_keymap[MWKEY_KP_ENTER&0xFF] = SDLK_KP_ENTER;
	MISC_keymap[MWKEY_KP_EQUALS&0xFF] = SDLK_KP_EQUALS;

	MISC_keymap[MWKEY_UP&0xFF] = SDLK_UP;
	MISC_keymap[MWKEY_DOWN&0xFF] = SDLK_DOWN;
	MISC_keymap[MWKEY_RIGHT&0xFF] = SDLK_RIGHT;
	MISC_keymap[MWKEY_LEFT&0xFF] = SDLK_LEFT;
	MISC_keymap[MWKEY_INSERT&0xFF] = SDLK_INSERT;
	MISC_keymap[MWKEY_DELETE&0xFF] = SDLK_DELETE;
	MISC_keymap[MWKEY_HOME&0xFF] = SDLK_HOME;
	MISC_keymap[MWKEY_END&0xFF] = SDLK_END;
	MISC_keymap[MWKEY_PAGEUP&0xFF] = SDLK_PAGEUP;
	MISC_keymap[MWKEY_PAGEDOWN&0xFF] = SDLK_PAGEDOWN;

	MISC_keymap[MWKEY_F1&0xFF] = SDLK_F1;
	MISC_keymap[MWKEY_F2&0xFF] = SDLK_F2;
	MISC_keymap[MWKEY_F3&0xFF] = SDLK_F3;
	MISC_keymap[MWKEY_F4&0xFF] = SDLK_F4;
	MISC_keymap[MWKEY_F5&0xFF] = SDLK_F5;
	MISC_keymap[MWKEY_F6&0xFF] = SDLK_F6;
	MISC_keymap[MWKEY_F7&0xFF] = SDLK_F7;
	MISC_keymap[MWKEY_F8&0xFF] = SDLK_F8;
	MISC_keymap[MWKEY_F9&0xFF] = SDLK_F9;
	MISC_keymap[MWKEY_F10&0xFF] = SDLK_F10;
	MISC_keymap[MWKEY_F11&0xFF] = SDLK_F11;
	MISC_keymap[MWKEY_F12&0xFF] = SDLK_F12;
	//MISC_keymap[XK_F13&0xFF] = SDLK_F13;
	//MISC_keymap[XK_F14&0xFF] = SDLK_F14;
	//MISC_keymap[XK_F15&0xFF] = SDLK_F15;

	MISC_keymap[MWKEY_NUMLOCK&0xFF] = SDLK_NUMLOCK;
	MISC_keymap[MWKEY_CAPSLOCK&0xFF] = SDLK_CAPSLOCK;
	MISC_keymap[MWKEY_SCROLLOCK&0xFF] = SDLK_SCROLLOCK;
	MISC_keymap[MWKEY_LSHIFT&0xFF] = SDLK_RSHIFT;
	MISC_keymap[MWKEY_RSHIFT&0xFF] = SDLK_LSHIFT;
	MISC_keymap[MWKEY_RCTRL&0xFF] = SDLK_RCTRL;
	MISC_keymap[MWKEY_LCTRL&0xFF] = SDLK_LCTRL;
	MISC_keymap[MWKEY_RALT&0xFF] = SDLK_RALT;
	MISC_keymap[MWKEY_LALT&0xFF] = SDLK_LALT;
	//MISC_keymap[MWKEY_RMETA&0xFF] = SDLK_RMETA;
	//MISC_keymap[MWKEY_LMETA&0xFF] = SDLK_LMETA;
	MISC_keymap[MWKEY_LMETA&0xFF] = SDLK_LSUPER; /* Left "Windows" */
	MISC_keymap[MWKEY_RMETA&0xFF] = SDLK_RSUPER; /* Right "Windows */
	MISC_keymap[MWKEY_ALTGR&0xFF] = SDLK_MODE; /* "Alt Gr" key */

	//MISC_keymap[XK_Clear&0xFF] = SDLK_CLEAR;
	MISC_keymap[MWKEY_PRINT&0xFF] = SDLK_PRINT;
	MISC_keymap[MWKEY_SYSREQ&0xFF] = SDLK_SYSREQ;
	MISC_keymap[MWKEY_PAUSE&0xFF] = SDLK_PAUSE;
	MISC_keymap[MWKEY_BREAK&0xFF] = SDLK_BREAK;
	MISC_keymap[MWKEY_MENU&0xFF] = SDLK_MENU;
	//MISC_keymap[XK_Help&0xFF] = SDLK_HELP;
	//MISC_keymap[XK_Hyper_R&0xFF] = SDLK_MENU;   /* Windows "Menu" key */
}

static SDL_keysym *FB_TranslateKey(int ch, int scancode, SDL_keysym *keysym)
{
	/* Set the keysym information */
	keysym->scancode = scancode;
	keysym->mod = KMOD_NONE;

	/* Get the translated SDL virtual keysym */
	if (ch & MWKEY_NONASCII_MASK) {
		keysym->sym = MISC_keymap[ch&0xFF];
	} else {
		switch (ch) {
		case '\b':
			keysym->sym = SDLK_BACKSPACE;
			break;
		case '\t':
			keysym->sym = SDLK_TAB;
			break;
		case '\r':
			keysym->sym = SDLK_RETURN;
			break;
		case 033:
			keysym->sym = SDLK_ESCAPE;
			break;
		default:
			keysym->sym = ch;
			break;
		}
	}

	/* If UNICODE is on, get the UNICODE value for the key */
	keysym->unicode = 0;
	if ( SDL_TranslateUNICODE )
		keysym->unicode = ch;
	return(keysym);
}

static int FB_DispatchEvent(_THIS, GR_EVENT *ev)
{
	int posted = 0;

	switch (ev->type) {
	    /* Mouse button press? */
	    case GR_EVENT_TYPE_BUTTON_DOWN: {
		posted = SDL_PrivateMouseButton(SDL_PRESSED, 
					ev->button.buttons, 0, 0);
	    }
	    break;

	    /* Mouse button release? */
	    case GR_EVENT_TYPE_BUTTON_UP: {
		posted = SDL_PrivateMouseButton(SDL_RELEASED, 
					ev->button.buttons, 0, 0);
	    }
	    break;

	    /* Key press? */
	    case GR_EVENT_TYPE_KEY_DOWN: {
		SDL_keysym keysym;
		posted = SDL_PrivateKeyboard(SDL_PRESSED,
				FB_TranslateKey(ev->keystroke.ch,
					ev->keystroke.scancode, &keysym));
	    }
	    break;

	    /* Key release? */
	    case GR_EVENT_TYPE_KEY_UP: {
		SDL_keysym keysym;

		/* Check to see if this is a repeated key */
		//if ( ! X11_KeyRepeat(SDL_Display, &xevent) ) {
			posted = SDL_PrivateKeyboard(SDL_RELEASED,
				FB_TranslateKey(ev->keystroke.ch,
					ev->keystroke.scancode, &keysym));
		//}
	    }
	    break;

	    /* Have we been requested to quit? */
	    case GR_EVENT_TYPE_CLOSE_REQ:
		posted = SDL_PrivateQuit();
		break;

	    case GR_EVENT_TYPE_UPDATE: {

	    	switch (ev->update.utype) {
		case GR_UPDATE_MAP:
			FB_SetWindowOffset(this, ev->update.x, ev->update.y);
			posted = SDL_PrivateAppActive(1, SDL_APPACTIVE);
			break;
		case GR_UPDATE_MOVE:
			FB_SetWindowOffset(this, ev->update.x, ev->update.y);
			break;
		case GR_UPDATE_UNMAP:
		case GR_UPDATE_UNMAPTEMP:
			posted = SDL_PrivateAppActive(0,
					SDL_APPACTIVE|SDL_APPINPUTFOCUS);
			FB_SetWindowOffset(this, -1, -1);
			break;
		case GR_UPDATE_SIZE:
			SDL_PrivateResize(ev->update.width, ev->update.height);
			break;
		}
	    }
	    break;

#if 0
	    /* Mouse motion? */
	    case MotionNotify: {
		if ( SDL_VideoSurface ) {
			if ( mouse_relative ) {
				if ( using_dga & DGA_MOUSE ) {
					posted = SDL_PrivateMouseMotion(0, 1,
							xevent.xmotion.x_root,
							xevent.xmotion.y_root);
				} else {
					posted = X11_WarpedMotion(this,&xevent);
				}
			} else {
				posted = SDL_PrivateMouseMotion(0, 0,
						xevent.xmotion.x,
						xevent.xmotion.y);
			}
		}
	    }
	    break;
	    /* Gaining mouse coverage? */
	    case EnterNotify: {
		if ( (xevent.xcrossing.mode != NotifyGrab) &&
		     (xevent.xcrossing.mode != NotifyUngrab) ) {
			posted = SDL_PrivateAppActive(1, SDL_APPMOUSEFOCUS);
		}
	    }
	    break;

	    /* Losing mouse coverage? */
	    case LeaveNotify: {
		if ( (xevent.xcrossing.mode != NotifyGrab) &&
		     (xevent.xcrossing.mode != NotifyUngrab) ) {
			posted = SDL_PrivateAppActive(0, SDL_APPMOUSEFOCUS);
		}
	    }
	    break;

	    /* Gaining input focus? */
	    case FocusIn: {
		posted = SDL_PrivateAppActive(1, SDL_APPINPUTFOCUS);

		/* Queue entry into fullscreen mode */
		switch_waiting = 0x01 | SDL_FULLSCREEN;
		switch_time = SDL_GetTicks() + 1500;
	    }
	    break;

	    /* Losing input focus? */
	    case FocusOut: {
		posted = SDL_PrivateAppActive(0, SDL_APPINPUTFOCUS);

		/* Queue leaving fullscreen mode */
		switch_waiting = 0x01;
		switch_time = SDL_GetTicks() + 200;
	    }
	    break;
	    /* Have we been restored? */
	    case MapNotify: {
		posted = SDL_PrivateAppActive(1, SDL_APPACTIVE);

		if ( SDL_VideoSurface &&
		     (SDL_VideoSurface->flags & SDL_FULLSCREEN) ) {
#ifdef GRAB_FULLSCREEN
			X11_EnterFullScreen(this);
#else
			/* Queue entry into fullscreen mode */
			switch_waiting = 0x01 | SDL_FULLSCREEN;
			switch_time = SDL_GetTicks() + 1500;
#endif
		} else {
			X11_GrabInputNoLock(this, this->input_grab);
		}
		if ( SDL_VideoSurface ) {
			X11_RefreshDisplay(this);
		}
	    }
	    break;

	    /* Do we need to refresh ourselves? */
	    case Expose: {
		if ( SDL_VideoSurface && (xevent.xexpose.count == 0) ) {
			X11_RefreshDisplay(this);
		}
	    }
	    break;

	    default: {
		/* Only post the event if we're watching for it */
		if ( SDL_ProcessEvents[SDL_SYSWMEVENT] == SDL_ENABLE ) {
			SDL_SysWMmsg wmmsg;

			SDL_VERSION(&wmmsg.version);
			wmmsg.subsystem = SDL_SYSWM_X11;
			wmmsg.event.xevent = xevent;
			posted = SDL_PrivateSysWMEvent(&wmmsg);
		}
	    }
	    break;
#endif
	    default:
	    	break;
	}
	return(posted);
}

void FB_PumpEvents(_THIS)
{
	GR_EVENT event;

	SDL_mutexP(hw_lock);
	do {
		GrCheckNextEvent(&event);
		FB_DispatchEvent(this, &event);
	} while (event.type != GR_EVENT_TYPE_NONE);
	SDL_mutexV(hw_lock);
}
